@@ -4,14 +4,13 @@
 #' @return A Huber mean estimator will be returned.
 #' @references Huber, P. J. (1964). Robust estimation of a location parameter. Ann. Math. Statist., 35, 73–101.
 #' @references Wang, L., Zheng, C., Zhou, W. and Zhou, W.-X. (2018). A New Principle for Tuning-Free Huber Regression. Preprint.
-#' @seealso \code{\link{farm.cov}} for tuning-free Huber-type covariance estimation.
+#' @seealso \code{\link{huber.cov}} for tuning-free Huber-type covariance estimation and \code{\link{huber.reg}} for tuning-free Huber regression.
 #' @examples
-#' set.seed(2019)
-#' n = 1000
-#' X = rlnorm(n, 0, 1.5) - exp(1.5^2 / 2)
-#' mu = farm.mean(X)
+#' n = 10000
+#' X = rt(n, 2) + 2
+#' mu = huber.mean(X)
 #' @export
-farm.mean = function(X){
+huber.mean = function(X){
   n = length(X)
   return (huberMean(X, n))
 }
@@ -22,21 +21,44 @@ farm.mean = function(X){
 #' @param X An \eqn{n} by \eqn{p} data matrix.
 #' @return A \eqn{p} by \eqn{p} Huber-type covariance matrix estimator will be returned.
 #' @references Huber, P. J. (1964). Robust estimation of a location parameter. Ann. Math. Statist., 35, 73–101.
-#' @references Ke, Y., Minsker, S., Ren, Z., Sun, Q. and Zhou, W.-X. (2019). User-friendly covariance estimation for heavy-tailed distributions: A survey and recent results. Statis. Sci., to appear.
-#' @seealso \code{\link{farm.mean}} for tuning-free Huber mean estimation.
+#' @references Ke, Y., Minsker, S., Ren, Z., Sun, Q. and Zhou, W.-X. (2019). User-friendly covariance estimation for heavy-tailed distributions. Statis. Sci., 34, 454-471.
+#' @seealso \code{\link{huber.mean}} for tuning-free Huber mean estimation and \code{\link{huber.reg}} for tuning-free Huber regression.
 #' @examples
-#' set.seed(2019)
 #' n = 100
 #' d = 50
 #' X = matrix(rt(n * d, df = 3), n, d) / sqrt(3)
-#' Sigma = farm.cov(X)
+#' Sigma = huber.cov(X)
 #' @export
-farm.cov = function(X) {
+huber.cov = function(X) {
   n = nrow(X)
   p = ncol(X)
   return (huberCov(X, n, p)$cov)
 }
 
+#' @title Tuning-free Huber regression
+#' @description The function calculates adaptive Huber regression estimator from a data sample, with robustification parameter \eqn{\tau} determined by a tuning-free principle.
+#' @param X An \eqn{n} by \eqn{p} design matrix, where \eqn{p < n}.
+#' @param Y A continuous response with length \eqn{n}.
+#' @return A coefficients estimator with length \eqn{p + 1} will be returned.
+#' @references Huber, P. J. (1964). Robust estimation of a location parameter. Ann. Math. Statist., 35, 73–101.
+#' @references Sun, Q., Zhou, W.-X. and Fan, J. (2020). Adaptive Huber regression. J. Amer. Statist. Assoc., 115, 254-265.
+#' @references Wang, L., Zheng, C., Zhou, W. and Zhou, W.-X. (2018). A new principle for tuning-free Huber regression. Preprint.
+#' @seealso \code{\link{huber.mean}} for tuning-free Huber mean estimation and \code{\link{huber.cov}} for tuning-free Huber-type covariance estimation.
+#' @examples
+#' n = 200
+#' d = 10
+#' beta = rep(1, d)
+#' X = matrix(rnorm(n * d), n, d)
+#' err = rnorm(n)
+#' Y = 1 + X %*% beta + err
+#' beta.hat = huber.reg(X, Y)
+#' @export
+huber.reg = function(X, Y) {
+  n = nrow(X)
+  p = ncol(X)
+  return (huberReg(X, Y, n, p))
+}
+
 #' @title Factor-adjusted robust multiple testing
 #' @description This function conducts factor-adjusted robust multiple testing (FarmTest) for means of multivariate data proposed in Fan et al. (2019) via a tuning-free procedure.
 #' @param X An \eqn{n} by \eqn{p} data matrix with each row being a sample.
@@ -48,46 +70,45 @@ farm.cov = function(X) {
 #' @param h0 An \strong{optional} \eqn{p}-vector of true means, or difference in means for two-sample FarmTest. The default is a zero vector.
 #' @param alternative An \strong{optional} character string specifying the alternate hypothesis, must be one of "two.sided" (default), "less" or "greater".
 #' @param alpha An \strong{optional} level for controlling the false discovery rate. The value of \code{alpha} must be between 0 and 1. The default value is 0.05.
-#' @param p.method An \strong{optional} character string specifying the method to calculate p-values when \code{fX} is known or when \code{KX = 0}, possible options are multiplier bootstrap or normal approximation. It must be one of "bootstrap" (default) or "normal".
+#' @param p.method An \strong{optional} character string specifying the method to calculate p-values when \code{fX} is known or when \code{KX = 0}, possible options are multiplier bootstrap or normal approximation. It must be one of "bootstrap"(default) or "normal".
 #' @param nBoot An \strong{optional} positive integer specifying the size of bootstrap sample, only available when \code{p.method = "bootstrap"}. The dafault value is 500.
 #' @return An object with S3 class \code{farm.test} containing the following items will be returned:
-#' \itemize{
-#' \item \code{means} Estimated means, a vector with length \eqn{p}.
-#' \item \code{stdDev} Estimated standard deviations, a vector with length \eqn{p}. It's not available for bootstrap method.
-#' \item \code{loadings} Estimated factor loadings, a matrix with dimension \eqn{p} by \eqn{K}, where \eqn{K} is the number of factors.
-#' \item \code{eigenVal} Eigenvalues of estimated covariance matrix, a vector with length \eqn{p}. It's only available when factors \code{fX} and \code{fY} are not given.
-#' \item \code{eigenRatio} Ratios of \code{eigenVal} to estimate \code{nFactors}, a vector with length \eqn{min(n, p) / 2}. It's only available when number of factors \code{KX} and \code{KY} are not given.
-#' \item \code{nFactors} Estimated or input number of factors, a positive integer.
-#' \item \code{tStat} Values of test statistics, a vector with length \eqn{p}. It's not available for bootstrap method.
-#' \item \code{pValues} P-values of tests, a vector with length \eqn{p}.
-#' \item \code{significant} Boolean values indicating whether each test is significant, with 1 for significant and 0 for non-significant, a vector with length \eqn{p}.
-#' \item \code{reject} Indices of tests that are rejected. It will show "no hypotheses rejected" if none of the tests are rejected.
-#' \item \code{type} Indicator of whether factor is known or unknown.
-#' \item \code{n} Sample size.
-#' \item \code{p} Data dimension.
-#' \item \code{h0} Null hypothesis, a vector with length \eqn{p}.
-#' \item \code{alpha} \eqn{\alpha} value.
-#' \item \code{alternative} Althernative hypothesis.
+#' \describe{
+#' \item{\code{means}}{Estimated means, a vector with length \eqn{p}.}
+#' \item{\code{stdDev}}{Estimated standard deviations, a vector with length \eqn{p}. It's not available for bootstrap method.}
+#' \item{\code{loadings}}{Estimated factor loadings, a matrix with dimension \eqn{p} by \eqn{K}, where \eqn{K} is the number of factors.}
+#' \item{\code{eigenVal}}{Eigenvalues of estimated covariance matrix, a vector with length \eqn{p}. It's only available when factors \code{fX} and \code{fY} are not given.}
+#' \item{\code{eigenRatio}}{Ratios of \code{eigenVal} to estimate \code{nFactors}, a vector with length \eqn{min(n, p) / 2}. It's only available when number of factors \code{KX} and \code{KY} are not given.}
+#' \item{\code{nFactors}}{Estimated or input number of factors, a positive integer.}
+#' \item{\code{tStat}}{Values of test statistics, a vector with length \eqn{p}. It's not available for bootstrap method.}
+#' \item{\code{pValues}}{P-values of tests, a vector with length \eqn{p}.}
+#' \item{\code{pAdjust}}{Adjusted p-values of tests, a vector with length \eqn{p}.}
+#' \item{\code{significant}}{Boolean values indicating whether each test is significant, with 1 for significant and 0 for non-significant, a vector with length \eqn{p}.}
+#' \item{\code{reject}}{Indices of tests that are rejected. It will show "no hypotheses rejected" if none of the tests are rejected.}
+#' \item{\code{type}}{Indicator of whether factor is known or unknown.}
+#' \item{\code{n}}{Sample size.}
+#' \item{\code{p}}{Data dimension.}
+#' \item{\code{h0}}{Null hypothesis, a vector with length \eqn{p}.}
+#' \item{\code{alpha}}{\eqn{\alpha} value.}
+#' \item{\code{alternative}}{Althernative hypothesis.}
 #' }
 #' @details For two-sample FarmTest, \code{means}, \code{stdDev}, \code{loadings}, \code{eigenVal}, \code{eigenRatio}, \code{nfactors} and \code{n} will be lists of items for sample X and Y separately.
 #' @details \code{alternative = "greater"} is the alternative that \eqn{\mu > \mu_0} for one-sample test or \eqn{\mu_X > \mu_Y} for two-sample test.
 #' @details Setting \code{p.method = "bootstrap"} for factor-known model will slow down the program, but it will achieve lower empirical FDP than setting \code{p.method = "normal"}.
 #' @references Ahn, S. C. and Horenstein, A. R. (2013). Eigenvalue ratio test for the number of factors. Econometrica, 81(3) 1203–1227.
 #' @references Benjamini, Y. and Hochberg, Y. (1995). Controlling the false discovery rate: A practical and powerful approach to multiple testing. J. R. Stat. Soc. Ser. B. Stat. Methodol., 57 289–300.
-#' @references Fan, J., Ke, Y., Sun, Q. and Zhou, W-X. (2019). FarmTest: Factor-adjusted robust multiple testing with approximate false discovery control. J. Amer. Statist. Assoc., to appear.
+#' @references Fan, J., Ke, Y., Sun, Q. and Zhou, W-X. (2019). FarmTest: Factor-adjusted robust multiple testing with approximate false discovery control. J. Amer. Statist. Assoc., 114, 1880-1893.
 #' @references Huber, P. J. (1964). Robust estimation of a location parameter. Ann. Math. Statist., 35, 73–101.
 #' @references Storey, J. D. (2002). A direct approach to false discovery rates. J. R. Stat. Soc. Ser. B. Stat. Methodol., 64, 479–498.
-#' @references Sun, Q., Zhou, W.-X. and Fan, J. (2019). Adaptive Huber regression. J. Amer. Statist. Assoc., to appear.
+#' @references Sun, Q., Zhou, W.-X. and Fan, J. (2020). Adaptive Huber regression. J. Amer. Statist. Assoc., 115, 254-265.
 #' @references Zhou, W-X., Bose, K., Fan, J. and Liu, H. (2018). A new perspective on robust M-estimation: Finite sample theory and applications to dependence-adjusted multiple testing. Ann. Statist., 46 1904-1931.
-#' @seealso \code{\link{print.farm.test}}
-#' @seealso \code{\link{farm.fdr}}
+#' @seealso \code{\link{print.farm.test}}, \code{\link{summary.farm.test}} and \code{\link{plot.farm.test}}.
 #' @examples 
 #' n = 20
 #' p = 50
 #' K = 3
 #' muX = rep(0, p)
 #' muX[1:5] = 2
-#' set.seed(2019)
 #' epsilonX = matrix(rnorm(p * n, 0, 1), nrow = n)
 #' BX = matrix(runif(p * K, -2, 2), nrow = p)
 #' fX = matrix(rnorm(K * n, 0, 1), nrow = n)
@@ -146,8 +167,9 @@ farm.test = function(X, fX = NULL, KX = -1, Y = NULL, fY = NULL, KY = -1, h0 = N
         reject = which(rst.list$significant == 1)
       }
       output = list(means = rst.list$means, stdDev = stdDev, loadings = loadings, eigenVal = eigenVal, eigenRatio = eigenRatio, 
-                    nFactors = rst.list$nfactors, tStat = tStat, pValues = rst.list$pValues, significant = rst.list$significant, 
-                    reject = reject, type = "known", n = nrow(X), p = p, h0 = h0, alpha = alpha, alternative = alternative)
+                    nFactors = rst.list$nfactors, tStat = tStat, pValues = rst.list$pValues, pAdjust = rst.list$pAdjust, 
+                    significant = rst.list$significant, reject = reject, type = "known", n = nrow(X), p = p, h0 = h0, alpha = alpha, 
+                    alternative = alternative)
     }
   } else if (is.null(Y) && is.null(fX)) {
     if (KX > p) {
@@ -169,8 +191,8 @@ farm.test = function(X, fX = NULL, KX = -1, Y = NULL, fY = NULL, KY = -1, h0 = N
         reject = which(rst.list$significant == 1)
       }
       output = list(means = rst.list$means, stdDev = stdDev, loadings = loadings, eigenVal = eigenVal, eigenRatio = eigenRatio, nFactors = 0, 
-                    tStat = tStat, pValues = rst.list$pValues, significant = rst.list$significant, reject = reject, type = "unknown", 
-                    n = nrow(X), p = p, h0 = h0, alpha = alpha, alternative = alternative)
+                    tStat = tStat, pValues = rst.list$pValues, pAdjust = rst.list$pAdjust, significant = rst.list$significant, reject = reject, 
+                    type = "unknown", n = nrow(X), p = p, h0 = h0, alpha = alpha, alternative = alternative)
     } else {
       eigenRatio = "not available when KX is specified"
       rst.list = farmTest(X, h0, KX, alpha, alternative)
@@ -182,8 +204,8 @@ farm.test = function(X, fX = NULL, KX = -1, Y = NULL, fY = NULL, KY = -1, h0 = N
       }
       output = list(means = rst.list$means, stdDev = rst.list$stdDev, loadings = rst.list$loadings, eigenVal = rst.list$eigens, 
                     eigenRatio = eigenRatio, nFactors = rst.list$nfactors, tStat = rst.list$tStat, pValues = rst.list$pValues, 
-                    significant = rst.list$significant, reject = reject, type = "unknown", n = nrow(X), p = p, h0 = h0, alpha = alpha, 
-                    alternative = alternative)
+                    pAdjust = rst.list$pAdjust, significant = rst.list$significant, reject = reject, type = "unknown", n = nrow(X), p = p, h0 = h0, 
+                    alpha = alpha, alternative = alternative)
     }
   } else if (!is.null(Y) && !is.null(fX)) {
     if (ncol(X) != ncol(Y)) {
@@ -215,8 +237,8 @@ farm.test = function(X, fX = NULL, KX = -1, Y = NULL, fY = NULL, KY = -1, h0 = N
       nfactors = list(X.nFactors = rst.list$nfactorsX, Y.nFactors = rst.list$nfactorsY)
       n = list(X.n = nrow(X), Y.n = nrow(Y))
       output = list(means = means, stdDev = stdDev, loadings = loadings, eigenVal = eigenVal, eigenRatio = eigenRatio, nFactors = nfactors, 
-                    tStat = tStat, pValues = rst.list$pValues, significant = rst.list$significant, reject = reject, type = "known", n = n, 
-                    p = p, h0 = h0, alpha = alpha, alternative = alternative)
+                    tStat = tStat, pValues = rst.list$pValues, pAdjust = rst.list$pAdjust, significant = rst.list$significant, reject = reject, 
+                    type = "known", n = n, p = p, h0 = h0, alpha = alpha, alternative = alternative)
     }
   } else {
     if (ncol(X) != ncol(Y)) {
@@ -247,8 +269,8 @@ farm.test = function(X, fX = NULL, KX = -1, Y = NULL, fY = NULL, KY = -1, h0 = N
       nfactors = list(X.nFactors = 0, Y.nFactors = 0)
       n = list(X.n = nrow(X), Y.n = nrow(Y))
       output = list(means = means, stdDev = stdDev, loadings = loadings, eigenVal = eigenVal, eigenRatio = eigenRatio, nFactors = nfactors, 
-                    tStat = tStat, pValues = rst.list$pValues, significant = rst.list$significant, reject = reject, type = "unknown", n = n, 
-                    p = p, h0 = h0, alpha = alpha, alternative = alternative)
+                    tStat = tStat, pValues = rst.list$pValues, pAdjust = rst.list$pAdjust, significant = rst.list$significant, reject = reject, 
+                    type = "unknown", n = n, p = p, h0 = h0, alpha = alpha, alternative = alternative)
     } else {
       rst.list = farmTestTwo(X, Y, h0, KX, KY, alpha, alternative)
       if (sum(rst.list$significant) > 0) {
@@ -270,33 +292,32 @@ farm.test = function(X, fX = NULL, KX = -1, Y = NULL, fY = NULL, KY = -1, h0 = N
       }
       eigenRatio = list(X.eigenRatio = ratioX, Y.eigenRatio = ratioY)
       output = list(means = means, stdDev = stdDev, loadings = loadings, eigenVal = eigenVal, eigenRatio = eigenRatio, nFactors = nfactors, 
-                    tStat = rst.list$tStat, pValues = rst.list$pValues, significant = rst.list$significant, reject = reject, type = "unknown", 
-                    n = n, p = p, h0 = h0, alpha = alpha, alternative = alternative)
+                    tStat = rst.list$tStat, pValues = rst.list$pValues, pAdjust = rst.list$pAdjust, significant = rst.list$significant, 
+                    reject = reject, type = "unknown", n = n, p = p, h0 = h0, alpha = alpha, alternative = alternative)
     } 
   }
   attr(output, "class") = "farm.test"
   return (output)
 }
 
-#' @title Summary and print function of FarmTest
-#' @description This is the print function of S3 objects with class "\code{farm.test}". It summarizes and prints the outputs of \code{farm.test} function.
+#' @title Print function of FarmTest
+#' @description This is the print function of S3 objects with class "\code{farm.test}".
 #' @param x A \code{farm.test} object.
 #' @param \dots Further arguments passed to or from other methods.
 #' @return No variable will be returned, but a brief summary of FarmTest will be displayed.
-#' @seealso \code{\link{farm.test}}
+#' @seealso \code{\link{farm.test}}, \code{\link{summary.farm.test}} and \code{\link{plot.farm.test}}.
 #' @examples 
 #' n = 50
 #' p = 100
 #' K = 3
 #' muX = rep(0, p)
 #' muX[1:5] = 2
-#' set.seed(2019)
 #' epsilonX = matrix(rnorm(p * n, 0, 1), nrow = n)
 #' BX = matrix(runif(p * K, -2, 2), nrow = p)
 #' fX = matrix(rnorm(K * n, 0, 1), nrow = n)
 #' X = rep(1, n) %*% t(muX) + fX %*% t(BX) + epsilonX
 #' output = farm.test(X)
-#' output
+#' print(output)
 #' @export
 print.farm.test = function(x, ...) {
   if (x$type == "known" && length(x$n) == 1) {
@@ -327,34 +348,63 @@ print.farm.test = function(x, ...) {
   cat(paste("Number of hypotheses rejected: ", sum(x$significant), "\n", sep = ""))
 }
 
-#' @title FDR control given a sequence of p-values
-#' @description Given a sequence of p-values, this function conducts multiple testing and outputs the indices of rejected hypotheses using an adaptive Benjamini-Hochberg procedure proposed by Storey (2002).
-#' @param pValues A sequence of p-values, each entry of \code{pValues} must be between 0 and 1.
-#' @param alpha An \strong{optional} level for controlling the false discovery rate. The value of \code{alpha} must be strictly between 0 and 1. The default value is 0.05.
-#' @return Indices of tests that are rejected will be returned. It will show "no hypotheses rejected" if none of the tests are rejected.
-#' @references Benjamini, Y. and Hochberg, Y. (1995). Controlling the false discovery rate: A practical and powerful approach to multiple testing. J. R. Stat. Soc. Ser. B. Stat. Methodol., 57 289–300.
-#' @references Storey, J. D. (2002). A direct approach to false discovery rates. J. R. Stat. Soc. Ser. B. Stat. Methodol., 64, 479–498.
+#' @title Summary function of FarmTest
+#' @description This is the summary function of S3 objects with class "\code{farm.test}".
+#' @param object A \code{farm.test} object.
+#' @param \dots Further arguments passed to or from other methods.
+#' @return A data frame including the estimated means, p-values, adjusted p-values and significance for all the features will be presented.
+#' @details For two-sample FarmTest, the first column is the difference: estimated means of sample \code{X} - estimated means of sample \code{Y}.
+#' @seealso \code{\link{farm.test}}, \code{\link{print.farm.test}} and \code{\link{plot.farm.test}}.
 #' @examples 
-#' set.seed(100)
 #' n = 50
 #' p = 100
-#' X = cbind(matrix(rnorm(n * 10, 1, 1), n, 10), matrix(rnorm(n * (p - 10)), n, p - 10))
-#' pValues = apply(X, 2, function(x) t.test(x)$p.value)
-#' farm.fdr(pValues)
-#' @seealso \code{\link{farm.test}}
+#' K = 3
+#' muX = rep(0, p)
+#' muX[1:5] = 2
+#' epsilonX = matrix(rnorm(p * n, 0, 1), nrow = n)
+#' BX = matrix(runif(p * K, -2, 2), nrow = p)
+#' fX = matrix(rnorm(K * n, 0, 1), nrow = n)
+#' X = rep(1, n) %*% t(muX) + fX %*% t(BX) + epsilonX
+#' output = farm.test(X)
+#' summary(output)
 #' @export
-farm.fdr = function(pValues, alpha = 0.05) {
-  if((sum(pValues < 0) != 0) || (sum(pValues > 1) != 0)) {
-    stop("Each p-value must be between 0 and 1")
-  } else if (alpha >= 1 || alpha <= 0) {
-    stop("Alpha should be strictly between 0 and 1")
+summary.farm.test = function(object, ...) {
+  rst = NULL
+  if (length(object$n) == 1) {
+    rst = as.data.frame(cbind(object$means, object$pValues, object$pAdjust, object$significant))
   } else {
-    p = length(pValues)
-    significant = getRej(pValues, alpha, p)
-    reject = "no hypotheses rejected"
-    if (sum(significant) > 0) {
-      reject = which(significant == 1)
-    }
-    return (reject)
+    rst = as.data.frame(cbind(object$means$X.means - object$means$Y.means, object$pValues, object$pAdjust, object$significant))
   }
+  names(rst) = c("means", "p-values", "p-adjusted", "significance")
+  return (rst)
+}
+
+#' @title Plot function of FarmTest
+#' @description This is the plot function of S3 objects with class "\code{farm.test}". It produces the histogram of estimated means.
+#' @param x A \code{farm.test} object.
+#' @param \dots Further arguments passed to or from other methods.
+#' @return No variable will be returned, but a histogram of estimated means will be presented.
+#' @details For two-sample FarmTest, the histogram is based on the difference: estimated means of sample \code{X} - estimated means of sample \code{Y}.
+#' @seealso \code{\link{farm.test}}, \code{\link{print.farm.test}} and \code{\link{summary.farm.test}}.
+#' @examples 
+#' n = 50
+#' p = 100
+#' K = 3
+#' muX = rep(0, p)
+#' muX[1:5] = 2
+#' epsilonX = matrix(rnorm(p * n, 0, 1), nrow = n)
+#' BX = matrix(runif(p * K, -2, 2), nrow = p)
+#' fX = matrix(rnorm(K * n, 0, 1), nrow = n)
+#' X = rep(1, n) %*% t(muX) + fX %*% t(BX) + epsilonX
+#' output = farm.test(X)
+#' plot(output)
+#' @export
+plot.farm.test = function(x, ...) {
+  means = NULL
+  if (length(x$n) == 1) {
+    means = as.vector(x$means)
+  } else {
+    means = as.vector(x$means$X.means - x$means$Y.means)
+  }
+  graphics::hist(means, freq = TRUE, main = "Histogram of Estimated Means", xlab = "Estimated Means", col = "blue")
 }
